from keras.initializers import Constant
from keras.callbacks import ModelCheckpoint
import sys, os, re, csv, codecs, numpy as np, pandas as pd
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences
from keras.layers import Dense, Input, LSTM, Embedding, Dropout, Activation
from keras.layers import Bidirectional, GlobalMaxPool1D,Bidirectional
from keras.models import Model
from keras import initializers, regularizers, constraints, optimizers, layers
from utils import get_data_splits, loadEmbeddingMatrix, vectorize, load_data, load_embeddings
import numpy as np
import pandas as pd
import gensim.models.keyedvectors as word2vec
import gc



def get_model(sentence_maxlen, embd_name, word_index):

	drop = 0.5

	embd_matrix, embd_dim = loadEmbeddingMatrix(embd_name, word_index) #load_glove_weights('./resources/glove', embd_dim, vocab_size, w2i)
	#embd_matrix, embd_dim = load_embeddings(word_index, sentence_maxlen, embd_name)

	# this returns a tensor
	print("Creating Model...")
	sequence_input = Input(shape=(sentence_maxlen,), dtype='int32')
	#embedding = Embedding(input_dim=vocabulary_size, output_dim=embedding_dim, input_length=sequence_length)(inputs)
	embedding_layer = Embedding(len(word_index) + 1,
							embd_dim,
							embeddings_initializer=Constant(embd_matrix),
							input_length=sentence_maxlen,
							trainable=False)(sequence_input)
	x = Bidirectional(LSTM(60, return_sequences=True,name='lstm_layer',dropout=0.1,recurrent_dropout=0.1))(embedding_layer)
	x = GlobalMaxPool1D()(x)
	x = Dropout(0.1)(x)
	x = Dense(50, activation="relu")(x)
	x = Dropout(0.1)(x)
	x = Dense(1, activation="relu")(x)
	model = Model(inputs=sequence_input, outputs=x)
	model.compile(loss='binary_crossentropy',
					  optimizer='adam',
					  metrics=['accuracy'])

	return model

def run_model(train,test,embd):
	np.random.seed(42)
	epochs=50
	# cut texts after this number of words
	# (among top max_features most common words)
	#GLOVE_DIR = "./resources/glove"
	embd_name = embd
	#embd_dim = 300
	train_data = load_data(train)
	test_data = load_data(test)

	data = train_data+test_data

	#train, dev, test = get_data_splits(df)
	#train_data = load_data(train)
	#dev_data = load_data(dev)
	#test_data = load_data(test)
	sentence_maxlen = max(map(len, (d for d, _ in data)))

	vocab = []
	for d, _ in data:
		for w in d:
			if w not in vocab: vocab.append(w)
	vocab = sorted(vocab)
	print('vocab size', len(vocab))
	w2i = {w:i for i,w in enumerate(vocab)}

	np.random.shuffle(data)
	X_train, Y_train = vectorize(train_data, sentence_maxlen, w2i)
	X_test, Y_test = vectorize(test_data, sentence_maxlen, w2i)

	print(X_train.shape)
	print(X_test.shape)
	epochs = 5
	batch_size = 30

	filepath="./models_chkpt/bilstm_best_{epoch:02d}-{val_acc:.2f}.hdf5"
	checkpoint = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
	callbacks_list = [checkpoint]

	model = get_model(sentence_maxlen, embd_name, w2i)
	print(model.summary())
	print("Traning Model...")

	model.fit(X_train, Y_train,
				batch_size=32,
				epochs=epochs,
				callbacks=callbacks_list,
				validation_data=(X_test, Y_test)
			)
	Y_pred = model.predict(X_test)
	return Y_pred #f1_score(Y_test, Y_pred), accuracy_score(Y_test, Y_pred)
