import os
import numpy as np
import pandas as pd
import argparse
import warnings
import time
import tqdm

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import f1_score, balanced_accuracy_score, accuracy_score
from sklearn.model_selection import train_test_split, cross_val_score, KFold, StratifiedKFold
from sklearn.preprocessing import LabelEncoder
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier
from sklearn.preprocessing import StandardScaler
from sklearn import svm
from cnn import run_model
from bilstm import run_model as run_bilstm
import random
import pickle
from utils import get_data_splits

from resources import DATA_PATH

from keras.preprocessing import sequence
from keras.preprocessing.text import Tokenizer
from keras.models import Sequential
from keras.layers import Dense, Dropout, Embedding, LSTM, Bidirectional
def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn
cue_phrases = ['because', 'meanwhile', 'but', 'on the other hand', 'anyway', 'and', 'then', 'so',
				'therefore', 'thereby', 'though', 'although', 'since', 'if', 'however', 'or', 'before', 'even',
				'except', 'only', 'especially', 'after'
]


from resources import BASE_DIR

#TODO
# 1. create train/dev split



def TfidfBaseline(train, test, classifier="logreg"):
	# ##### PREPROCESS DATA
	vectorizer = TfidfVectorizer()
	X_train, Y_train = vectorizer.fit_transform(train['sentence_text']), train['label']
	X_test = vectorizer.transform(test['sentence_text'])

	if classifier == "logreg":
		clf = LogisticRegression().fit(X_train, Y_train)
	elif classifier == "mlp":
		clf = MLPClassifier().fit(X_train, Y_train)
	y_pred = clf.predict(X_test)
	
	return y_pred

def readFeatures(fnames,wanted_num):
	features = None
	for fn in fnames:
		if 'specificity' in fn:
			fpath = os.path.join(BASE_DIR,'specificity','train_spec_scores.txt')
		elif 'politeness' in fn:
			fpath = os.path.join(BASE_DIR,'politeness','train_plt_scores.txt')
		elif 'length' in fn:
			fpath = os.path.join(BASE_DIR,'specificity','train_sents_length.txt')
		elif 'infersent' in fn:
			fpath = os.path.join(BASE_DIR,'infersent','train_infersent_w2v.pkl')
		else:
			print('Invalide feature name ', fn)
			exit(123)

		if 'infersent' in fn:
			scores = pickle.load(open(fpath,'rb'))
		else:
			ff = open(fpath,'r')
			if 'log' in fn:
				scores = [np.log(float(f)) for f in ff.readlines()]
			else:
				scores = [float(f) for f in ff.readlines()]
		assert len(scores) == wanted_num
		if features is None:
			features = np.array(scores).reshape(wanted_num,-1)
		else:
			features = np.concatenate((features,np.array(scores).reshape(wanted_num,-1)), axis=1)

	return features

def downSampling(idx_list, labels, rr=1.):
	np_num = len([l for l in labels if l==0])
	p_num = len(labels) - np_num
	ratio = p_num*rr/np_num
	new_idx = [ii  for ii in range(idx_list.shape[0]) if labels[ii]==1 or (labels[ii]==0 and np.random.random()<ratio)]
	return idx_list[(new_idx)], labels[(new_idx)]

def splitFeatures(full_feature_matrix, full, train, test):
	train_idx = []
	test_idx = []

	for i,entry in tqdm.tqdm(train.iterrows(),desc='building training features'):
		aid = entry['article_title']
		alist = full['article_title'].tolist()
		sidx = [ii for ii in range(len(alist)) if alist[ii]==aid ]
		idx_list = []
		for sid in sidx:
			if full.iloc[sid]['sentence_text'] == entry['sentence_text']: idx_list.append(sid)
		train_idx.append(idx_list[0])

	assert len(train_idx) == len(train['label'].tolist()) #- len(train_fail_idx)

	for i,entry in tqdm.tqdm(test.iterrows(), desc='building testing features'):
		aid = entry['article_title']
		alist = full['article_title'].tolist()
		sidx = [ii for ii in range(len(alist)) if alist[ii]==aid ]
		idx_list = []
		for sid in sidx:
			if full.iloc[sid]['sentence_text'] == entry['sentence_text']: idx_list.append(sid)
		test_idx.append(idx_list[0])

	assert len(test_idx) == len(test['label'].tolist()) #- len(train_fail_idx)

	train_idx, train_labels = downSampling(train_idx,train['label'].tolist())

	return full_feature_matrix[np.array(train_idx),:], train_labels, full_feature_matrix[np.array(test_idx), :]


def infersentFeatures(train,test,wanted_features,clf_type='logistic-regression'):
	full_df = pd.read_csv(DATA_PATH, sep='\t')
	full_df.fillna('', inplace=True)
	full_df = full_df[full_df['sentence_text'] != '']

	train.fillna('', inplace=True)
	ctrain = train[train['sentence_text'] != '']
	test.fillna('', inplace=True)
	ctest = test[test['sentence_text'] != '']

	full_feature = readFeatures(wanted_features,full_df.shape[0])
	train_feature, train_labels, test_feature = splitFeatures(full_feature, full_df, ctrain, ctest)
	if 'svc' in clf_type:
		clf = svm.LinearSVC()
	else:
		clf = LogisticRegression()

	clf.fit(train_feature,train_labels)
	return clf.predict(test_feature)


if __name__  == '__main__':
	##### Parameters

	parser = argparse.ArgumentParser(description="Sentence classification for propaganda detection")
	parser.add_argument('-m', '--models', help="The models to run", nargs="*", default="tfidf")
	parser.add_argument('-f', '--fold', help="The fold to use", default=0, type=int)
	parser.add_argument('-e', '--embd', help="Embeddings", default="glove")
	args = parser.parse_args()

	train,test = get_data_splits(args.fold)
	le = LabelEncoder().fit(train['label'])
	train['label'] = le.transform(train['label'])
	test['label'] = le.transform(test['label'])
	y_true = test['label']

	#train_sentences, train_label = train['sentence_text'], train['label']
	#test_sentences, test_label = test['sentence_text'], test['label']

	results = []
	label_types = list(set(train['label']))

	if "tfidf" in args.models:
		clf = "mlp"
		y_pred = TfidfBaseline(train,test,clf)
		results.append({'model': "tfidf-" + clf, 'y_pred': y_pred})

	if "random" in args.models:
		### random baseline; average over 100 repeats
		f1s_prop = []
		f1s_nonprop = []
		accs = []
		kfold = 100
		for _ in range(kfold):
			pred = le.transform([random.choice(le.classes_) for _ in range(test.shape[0])])
			f1s_prop.append(f1_score(y_true, pred, pos_label=1))
			f1s_nonprop.append(f1_score(y_true, pred, pos_label=0))
			accs.append(accuracy_score(y_true, pred))
		results.append({"model": "random", "f1_prop": np.mean(f1s_prop), "f1_nonprop": np.mean(f1s_nonprop), "accuracy": np.mean(accs)})
		#print('random baseline (f1) : {}'.format())

	### majority baseline
	if "majority" in args.models:
		for ll in label_types:
			print('label {} num: {}'.format(ll,train[train['label']==ll].shape[0]))
		y_pred = le.transform(['non-propaganda' for _ in range(test.shape[0])])
		results.append({'model': "majority", 'y_pred': y_pred})
		#results.append({"model": , "f1": f1_score(y_true,pred), 
		#	"accuracy": accuracy_score(y_true, pred)})
		#print('majority baseline (f1) : {}'.format())

	if "cnn" in args.models:
		y_pred = run_model(train,test,args.embd) #f1, acc = run_model(df)
		thresholds = []
		for thresh in np.arange(0.1, 0.501, 0.01):
			thresh = np.round(thresh, 2)
			res = f1_score(y_true, (y_pred > thresh).astype(int))
			thresholds.append([thresh, res])
			#print("F1 score at threshold {0} is {1}".format(thresh, res))
		thresholds.sort(key=lambda x: x[1], reverse=True)
		best_thresh = thresholds[0][0]
		print('best threshold:', best_thresh)
		y_pred = (y_pred > best_thresh).astype(int)
		results.append({"model": "cnn",  'y_pred': y_pred})#"f1": f1, "accuracy": acc})
	
	if "bilstm" in args.models:
		y_pred = run_bilstm(train,test,args.embd) #f1, acc = run_model(df)
		thresholds = []
		for thresh in np.arange(0.1, 0.501, 0.01):
			thresh = np.round(thresh, 2)
			res = f1_score(y_true, (y_pred > thresh).astype(int))
			thresholds.append([thresh, res])
		thresholds.sort(key=lambda x: x[1], reverse=True)
		best_thresh = thresholds[0][0]
		print('best threshold:', best_thresh)
		y_pred = (y_pred > best_thresh).astype(int)
		results.append({"model": "bilstm",  'y_pred': y_pred})#"f1": f1, "accuracy": acc})

	if 'infersent' in args.models:
		wanted_features = ['infersent','politeness','log_length']
		print('wanted_features: {}'.format(wanted_features))
		y_pred = infersentFeatures(train,test,wanted_features)
		results.append({"model": "infersent",  'y_pred': y_pred})#"f1": f1, "accuracy": acc})

	print('Finished experiments')
	final = []
	for res in results:
		print('Model:', res['model'])
		if 'y_pred' in res:
			f1_prop = f1_score(y_true, res['y_pred'], pos_label=1)
			f1_nonprop = f1_score(y_true, res['y_pred'], pos_label=0)
			acc = accuracy_score(y_true, res['y_pred'])
			final.append({"f1_prop": f1_prop, "f1_nonprop": f1_nonprop, 
				"model": res["model"], "accuracy": acc, "fold": str(args.fold)})
			print('F1 (propaganda):', f1_prop)
			print('F1 (non-propaganda):', f1_nonprop)
			print('Accuracy:', acc)
		else:
			res["fold"] = str(args.fold)
			final.append(res)
			print('F1 (propaganda):', res['f1_prop'])
			print('F1 (non-propaganda):', res['f1_nonprop'])
			print('Accuracy:', res['accuracy'])
		print()

	pd.DataFrame(final).to_csv('results' + str(time.asctime( time.localtime(time.time()))) + ".csv", index=False)
	exit(123)


