from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.neural_network import MLPClassifier
from cnn import get_model as get_model_cnn
from keras.models import model_from_json

#BERT

import csv
import os
import logging
import argparse
import random
from tqdm import tqdm, trange

import numpy as np
import pandas as pd
import torch


import time
import argparse
from keras.callbacks import ModelCheckpoint
from keras.models import Sequential, Model
from keras.layers import Dense, Dropout,Input,GlobalMaxPooling1D
from keras.layers import Embedding
from keras.layers import Conv1D, GlobalAveragePooling1D, MaxPooling1D
from utils import get_data, loadEmbeddingMatrix, vectorize, load_data, load_embeddings, preprocess_text, vectorize_without_labels
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score, accuracy_score
from sklearn.preprocessing import LabelEncoder
from keras.initializers import Constant
from keras.layers import Input, Embedding, Dense, Conv2D, MaxPool2D
from keras.layers import Reshape, Flatten, Concatenate, Dropout, SpatialDropout1D
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences
from keras.callbacks import Callback
from bilstm import get_model as get_model_bilstm

class F1_Callback(Callback):
	def __init__(self, test_data, test_label, weight_fp):
		self.test_data = test_data
		self.test_label = test_label
		self.f1 = -10000.0
		self.weight_fp = weight_fp

	def on_epoch_end(self, epoch, logs=None):
		probs = self.model.predict(self.test_data)
		new_f1 = f1_score(self.test_label, np.round(probs).astype(int))
		if self.f1 < new_f1:
			print('old f1:', self.f1, ', new f1:', new_f1)
			print('better f1 score! save model')
			self.model.save_weights(self.weight_fp)


def get_best_thresh(y_true, y_pred):
	thresholds = []
	for thresh in np.arange(0.1, 0.501, 0.01):
		thresh = np.round(thresh, 2)
		res = f1_score(y_true, (y_pred > thresh).astype(int))
		thresholds.append([thresh, res])
	thresholds.sort(key=lambda x: x[1], reverse=True)
	best_thresh = thresholds[0][0]
	return best_thresh

def submit_cnn():
	#def get_model(sentence_maxlen, embedding_matrix, embd_dim, word_index,s_dropout=0.4, dropout=0.1):
	preprocess = True
	downsample = False
	embd = "word2vec"
	epochs=1
	batch_size=32
	dropout=0.1
	s_dropout=0.4

	train, test = get_data("train"), get_data("test") #, get_data("test")
	print('test:', test)
	le = LabelEncoder().fit(train['label'])
	train['label'] = le.transform(train['label'])
	#test['label'] = le.transform(test['label'])
	#y_true = test['label']

	np.random.seed(42)

	if preprocess:
		train = preprocess_text(train)
		test = preprocess_text(test)

	train_data = load_data(train)
	test_data = load_data(test)

	data = train_data+test_data

	sentence_maxlen = max(map(len, (d for d, _ in data)))

	vocab = []
	for d, _ in data:
		for w in d:
			if w not in vocab: vocab.append(w)
	vocab = sorted(vocab)
	w2i = {w:i for i,w in enumerate(vocab)}

	np.random.shuffle(train_data)

	X_train, Y_train = vectorize(train_data, sentence_maxlen, w2i)
	 

	#X_train, X_dev, Y_train, Y_dev = train_test_split(X_train, Y_train, test_size=0.2, stratify=Y_train)
	
	if downsample:
		X_train, Y_train = downSampling(X_train, Y_train)
	print("Loading Embeddings...")
	#embd_matrix, embd_dim = loadEmbeddingMatrix(embd, w2i) #load_glove_weights('./resources/glove', embd_dim, vocab_size, w2i)
	embd_matrix, embd_dim =np.zeros((len(w2i), 300)), 300
	
	strtime = str(time.time())
	print('weights:', strtime)
	#filepath="./models_chkpt/" + args.embd + "_cnn_best_{epoch:02d}-{val_acc:.2f}.hdf5"
	filepath = "./models_chkpt/" + strtime + "_weights.hdf5"
	checkpoint = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
	#checkpoint2 = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
	#checkpoint = F1_Callback(X_dev, Y_dev, filepath)
	callbacks_list = [checkpoint]
 
	model = get_model_cnn(sentence_maxlen, embd_matrix, embd_dim, w2i, dropout=dropout, s_dropout=s_dropout)
	print("Traning Model...")
	model.fit(X_train, Y_train,batch_size=batch_size,epochs=epochs,callbacks=callbacks_list) #, validation_data=(X_dev, Y_dev))
	
	# get best performing model
	#model.load_weights(filepath)

	#best_thresh = get_best_thresh(Y_dev, model.predict(X_dev))
	#print('Best threshold on dev:', best_thresh)

	print("Predicting")
	Y_pred = model.predict(X_test)
	print(len(Y_pred))
	Y_pred = np.round(Y_pred).astype(int)
	print(len(Y_pred))
	Y_pred = le.inverse_transform(np.ravel(Y_pred))
	print(len(Y_pred))
	print(Y_pred)
	test['label'] = Y_pred

	test.to_csv(embd + '_cnn_submission_predictions_test.txt', columns=['article_id', 'sentence_id', 'label'], index=False, sep="\t", header=False)
	#model = get_model_cnn()
	#loaded_json = open(model_json, 'r').read()
	#model = model_from_json(loaded_json)
	#model.load_weights(filepath)

def submit_bilstm():
	#def get_model(sentence_maxlen, embedding_matrix, embd_dim, word_index,s_dropout=0.4, dropout=0.1):
	preprocess = False
	downsample = False
	embd = "word2vec"
	epochs=2
	batch_size=32
	dropout=0.1
	s_dropout=0.4

	train, test = get_data("train"), get_data("dev") #, get_data("test")
	
	le = LabelEncoder().fit(train['label'])
	train['label'] = le.transform(train['label'])
	#test['label'] = le.transform(test['label'])
	#y_true = test['label']

	np.random.seed(42)

	train_data = load_data(train)
	test_data = load_data(test)

	if preprocess:
		train = preprocess_text(train)
		train = train[train['sentence_text'] != '']
		test = preprocess_text(test)
		test = test[test['sentence_text'] != '']

	data = train_data+test_data

	sentence_maxlen = max(map(len, (d for d, _ in data)))

	vocab = []
	for d, _ in data:
		for w in d:
			if w not in vocab: vocab.append(w)
	vocab = sorted(vocab)
	w2i = {w:i for i,w in enumerate(vocab)}

	np.random.shuffle(train_data)

	X_train, Y_train = vectorize(train_data, sentence_maxlen, w2i)
	X_test = vectorize_without_labels(test_data, sentence_maxlen, w2i)

	X_train_, X_dev_, Y_train, Y_dev = train_test_split(X_train, Y_train, test_size=0.2, stratify=Y_train)
	
	if downsample:
		X_train, Y_train = downSampling(X_train, Y_train)
	print("Loading Embeddings...")
	embd_matrix, embd_dim = loadEmbeddingMatrix(embd, w2i) #load_glove_weights('./resources/glove', embd_dim, vocab_size, w2i)
	#embd_matrix, embd_dim =np.zeros((len(w2i), 300)), 300
	
	strtime = str(time.time())
	#filepath="./models_chkpt/" + args.embd + "_cnn_best_{epoch:02d}-{val_acc:.2f}.hdf5"
	filepath = "./models_chkpt/" + strtime + "_weights.hdf5"
	checkpoint = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
	#checkpoint2 = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
	#checkpoint = F1_Callback(X_dev, Y_dev, filepath)
	callbacks_list = [checkpoint]

	model = get_model_bilstm(sentence_maxlen, embd_matrix, embd_dim, w2i, dropout=dropout, s_dropout=s_dropout)
	print("Traning Model...")
	model.fit(X_train, Y_train,batch_size=batch_size,epochs=epochs,callbacks=callbacks_list, validation_data=(X_dev, Y_dev))

	
	# get best performing model
	model.load_weights(filepath)
	print("Predicting")
	Y_pred = model.predict(X_test)
	Y_pred = np.round(Y_pred).astype(int)
	Y_pred = le.inverse_transform(np.ravel(Y_pred))
	test['label'] = Y_pred

	test.to_csv('bilstm_submission_predictions.txt', columns=['article_id', 'sentence_id', 'label'], index=False, sep="\t", header=False)

def submit_mlp():
	# ##### PREPROCESS DATA
	train, test = get_data("train"), get_data("dev") #, get_data("test")

	sentences = train['sentence_text'].tolist() + test['sentence_text'].tolist()

	vectorizer = TfidfVectorizer()
	vectorizer.fit(sentences)
	X_train, Y_train = vectorizer.fit_transform(train['sentence_text']), train['label']
	X_test = vectorizer.transform(test['sentence_text'])

	clf = MLPClassifier().fit(X_train, Y_train)
	y_pred = clf.predict(X_test)
	#y_pred = np.round(y_pred).astype(int)	
	test['label'] = y_pred
	test.to_csv('mlp_submission_predictions.txt', columns=['article_id', 'sentence_id', 'label'], index=False, sep="\t")

if __name__ == '__main__':
	parser = argparse.ArgumentParser(description="Sentence classification for propaganda detection")
	parser.add_argument('--model', help="The model to run", default="cnn", type=str)
	args = parser.parse_args()
	if args.model == "cnn":
		submit_cnn()
	if args.model == "bilstm":
		submit_bilstm()
	if args.model == "mlp":
		submit_mlp()
