import os.path
import glob
import os
import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder
import gensim.models.keyedvectors as word2vec
import gc
import nltk
nltk.download('stopwords')
from nltk.corpus import stopwords

def load_data(df):
	data = []
	for row in df.itertuples():
		data.append((row.sentence_text.split(' '), row.label))
	return data

def vectorize(data, sentence_maxlen, w2i):
	vec_data = []
	labels = []
	for d, label in data:
		vec = [w2i[w] for w in d if w in w2i]
		pad_len = max(0, sentence_maxlen - len(vec))
		vec += [0] * pad_len
		vec_data.append(vec)
		
		labels.append(label)
	vec_data = np.array(vec_data)
	labels = np.array(labels)
	return vec_data, labels

def vectorize_without_labels(data, sentence_maxlen, w2i):
	vec_data = []
	for d, label in data:
		vec = [w2i[w] for w in d if w in w2i]
		pad_len = max(0, sentence_maxlen - len(vec))
		vec += [0] * pad_len
		vec_data.append(vec)
		
	vec_data = np.array(vec_data)
	return vec_data

def parse_data():
	folder = "./datasets-v4/datasets-v4/tasks-2-3/train"
	sentences = []
	for filepath in glob.glob(folder + '/*.txt'):
		article_id = filepath.split('/')[-1].replace('.txt', '').replace('article', '')
		labels_path = filepath.replace(".txt", ".task2.labels")
		labels = []
		for line in open(labels_path):
			str_label = line.strip().split('\t')[-1]
			labels.append(str_label)
		with open(filepath, 'r') as f:
			content = f.read().split('\n')
		article_title = content[0]
		for idx, line in enumerate(open(filepath)):
			line = line.strip()
			if len(line) > 0:
				sentences.append({"article_id": article_id, "article_title": article_title, "sentence_text": line, "label": labels[idx]})
	return sentences

def preprocess_text(df):
	contraction_mapping = {"ain't": "is not", "aren't": "are not","can't": "cannot", "'cause": "because", "could've": "could have", "couldn't": "could not", "didn't": "did not",  "doesn't": "does not", "don't": "do not", "hadn't": "had not", "hasn't": "has not", "haven't": "have not", "he'd": "he would","he'll": "he will", "he's": "he is", "how'd": "how did", "how'd'y": "how do you", "how'll": "how will", "how's": "how is",  "I'd": "I would", "I'd've": "I would have", "I'll": "I will", "I'll've": "I will have","I'm": "I am", "I've": "I have", "i'd": "i would", "i'd've": "i would have", "i'll": "i will",  "i'll've": "i will have","i'm": "i am", "i've": "i have", "isn't": "is not", "it'd": "it would", "it'd've": "it would have", "it'll": "it will", "it'll've": "it will have","it's": "it is", "let's": "let us", "ma'am": "madam", "mayn't": "may not", "might've": "might have","mightn't": "might not","mightn't've": "might not have", "must've": "must have", "mustn't": "must not", "mustn't've": "must not have", "needn't": "need not", "needn't've": "need not have","o'clock": "of the clock", "oughtn't": "ought not", "oughtn't've": "ought not have", "shan't": "shall not", "sha'n't": "shall not", "shan't've": "shall not have", "she'd": "she would", "she'd've": "she would have", "she'll": "she will", "she'll've": "she will have", "she's": "she is", "should've": "should have", "shouldn't": "should not", "shouldn't've": "should not have", "so've": "so have","so's": "so as", "this's": "this is","that'd": "that would", "that'd've": "that would have", "that's": "that is", "there'd": "there would", "there'd've": "there would have", "there's": "there is", "here's": "here is","they'd": "they would", "they'd've": "they would have", "they'll": "they will", "they'll've": "they will have", "they're": "they are", "they've": "they have", "to've": "to have", "wasn't": "was not", "we'd": "we would", "we'd've": "we would have", "we'll": "we will", "we'll've": "we will have", "we're": "we are", "we've": "we have", "weren't": "were not", "what'll": "what will", "what'll've": "what will have", "what're": "what are",  "what's": "what is", "what've": "what have", "when's": "when is", "when've": "when have", "where'd": "where did", "where's": "where is", "where've": "where have", "who'll": "who will", "who'll've": "who will have", "who's": "who is", "who've": "who have", "why's": "why is", "why've": "why have", "will've": "will have", "won't": "will not", "won't've": "will not have", "would've": "would have", "wouldn't": "would not", "wouldn't've": "would not have", "y'all": "you all", "y'all'd": "you all would","y'all'd've": "you all would have","y'all're": "you all are","y'all've": "you all have","you'd": "you would", "you'd've": "you would have", "you'll": "you will", "you'll've": "you will have", "you're": "you are", "you've": "you have" }
	def clean_contractions(text, mapping):
		specials = ["’", "‘", "´", "`"]
		for s in specials:
			text = text.replace(s, "'")
		text = ' '.join([mapping[t] if t in mapping else t for t in text.split(" ")])
		return text
	punct = "1234567890/-'.?!,#$%\'()*+-/:;<=>@[\\]^_`{|}~" + '""“”’' + '∞θ÷α•à−β∅³π‘₹´°£€\×™√²—–&'
	def clean_punct(text, punct):
		for p in punct:
			text = text.replace(p, '')
		return text
	def clean_stopwords(text, stopwords):
		for stw in stopwords:
			text = text.replace(stw, '')
		return text
	df["sentence_text"] = df["sentence_text"].apply(lambda x: clean_contractions(x, contraction_mapping))
	df["sentence_text"] = df["sentence_text"].apply(lambda x: clean_punct(x, punct))
	df["sentence_text"] = df["sentence_text"].apply(lambda x: clean_stopwords(x, stopwords.words('english')))

	return df

def create_data_splits():
	"""
	Split the data in train/dev/test sets according to provided numbers
	We split it using the article title
	"""
	train = 0.8
	folds = 3

	##### READ DATA
	df = pd.read_csv('data.csv', sep='\t')
	df.fillna('', inplace=True)
	# remove empty sentences
	df = df[df['sentence_text'] != '']

	article_titles = pd.unique(df["article_title"])
	train_count = round(train * len(article_titles))

	for i in range(folds):
		np.random.shuffle(article_titles)
		train_articles = article_titles[:train_count]
		test_articles = article_titles[train_count:]
		#print(train_articles[:5])
		train_df = pd.DataFrame(df[df["article_title"].isin(train_articles)])
		test_df = pd.DataFrame(df[df["article_title"].isin(test_articles)])
		#print(sorted(pd.unique(train_df['article_title']))[:5])
		train_df.to_csv('./dataset/fold' + str(i) + '/train.tsv', sep="\t", index=False)
		test_df.to_csv('./dataset/fold' + str(i) + '/test.tsv', sep="\t", index=False)

def get_data_splits(fold=0):

	train = pd.read_csv('./dataset/fold' + str(fold) + '/train.tsv', sep="\t")
	test = pd.read_csv('./dataset/fold' + str(fold) + '/test.tsv', sep="\t")

	return train,test

def get_data(data_type):
	if data_type == "train":
		folder = "dataset/train/" # if train and dev folders are not in the same folder as this notebook, change
	elif data_type == "dev":
		folder = "dataset/dev/"
	elif data_type == "test":
		folder = "dataset/test/"
	# loading articles' content from *.txt files in the train folder
	file_list = glob.glob(os.path.join(folder, "*.txt"))
	sentence_list = []
	for i, filename in enumerate(file_list):
	    with open(filename, "r", encoding="utf-8") as f:
	        for row in f.readlines():
	            sentence_list.append(row.rstrip())

	# loading articles ids and sentence ids from files *.task2.labels in the train folder 
	gold_file_list = glob.glob(os.path.join(folder, "*.task2.labels"))
	articles_id, sentence_id_list, gold_labels = ([], [], [])
	for filename in gold_file_list:
	    with open(filename, "r") as f:
	        for row in f.readlines():
	            article_id, sentence_id, gold_label = row.rstrip().split("\t")
	            articles_id.append(article_id)
	            sentence_id_list.append(sentence_id)
	            gold_labels.append(gold_label)
	print(data_type + ": Loaded %d sentences from %d articles" % (len(sentence_list), i+1))
	return pd.DataFrame({"article_id": articles_id, 'sentence_id': sentence_id_list, 'label': gold_labels, 'sentence_text': sentence_list})

def loadEmbeddingMatrix(typeToLoad, word_index):
	if(typeToLoad=="glove"):
		EMBEDDING_FILE='./resources/glove/glove.840B.300d.txt'
		embed_size = 300
	elif(typeToLoad=="word2vec"):
		word2vecDict = word2vec.KeyedVectors.load_word2vec_format("./resources/GoogleNews-vectors-negative300.bin", binary=True)
		embed_size = 300
	elif(typeToLoad=="urban_glove"):
		EMBEDDING_FILE='./resources/urban_glove300.txt'
		embed_size=300
	elif(typeToLoad=="fasttext"):
		EMBEDDING_FILE='../input/fasttext/wiki.simple.vec'
		embed_size = 300

	if(typeToLoad=="glove" or typeToLoad=="fasttext"):
		embeddings_index = dict()
		f = open(EMBEDDING_FILE)
		for line in f:
			#split up line into an indexed array
			values = line.split(' ')
			#first index is word
			word = values[0]
			#store the rest of the values in the array as a new array
			
			coefs = np.asarray(values[1:], dtype='float32')
			embeddings_index[word] = coefs #50 dimensions
		f.close()
		print('Loaded %s word vectors.' % len(embeddings_index))
	elif(typeToLoad=="urban_glove"):
		embeddings_index = dict()
		f = open(EMBEDDING_FILE)
		for line in f:
			values = line.split(' ')
			word = values[0]
			coefs = np.asarray(values[1:], dtype='float32')
			embeddings_index[word] = coefs #50 dimensions
		f.close()
		nb_words = len(word_index)
		embedding_matrix = np.zeros((nb_words, embed_size))
		for word, i in word_index.items():
			embedding_vector = embeddings_index.get(word)
			if embedding_vector is not None: 
				embedding_matrix[i] = embedding_vector
		return embedding_matrix, embed_size
	else:
		embeddings_index = dict()
		for word in word2vecDict.wv.vocab:
			embeddings_index[word] = word2vecDict.word_vec(word)
		print('Loaded %s word vectors.' % len(embeddings_index))
		
	gc.collect()
	#We get the mean and standard deviation of the embedding weights so that we could maintain the 
	#same statistics for the rest of our own random generated weights.
	#print(embeddings_index.values()) 
	all_embs = np.stack(list(embeddings_index.values()))
	emb_mean,emb_std = all_embs.mean(), all_embs.std()
	
	nb_words = len(word_index)
	#We are going to set the embedding size to the pretrained dimension as we are replicating it.
	#the size will be Number of Words in Vocab X Embedding Size
	embedding_matrix = np.random.normal(emb_mean, emb_std, (nb_words, embed_size))
	gc.collect()

	#With the newly created embedding matrix, we'll fill it up with the words that we have in both 
	#our own dictionary and loaded pretrained embedding. 
	for word, i in word_index.items():
		#then we see if this word is in glove's dictionary, if yes, get the corresponding weights
		embedding_vector = embeddings_index.get(word)
		#and store inside the embedding matrix that we will train later on.
		if embedding_vector is not None: 
			embedding_matrix[i] = embedding_vector
	#print('total embedded:',embeddedCount,'common words')
	
	del(embeddings_index)
	gc.collect()
	
	#finally, return the embedding matrix
	return embedding_matrix, embed_size

def load_embeddings(word_index, types=["glove"]):
	print('loading embeddings:', types)
	embeddings = []
	if "glove" in types:
		EMBEDDING_FILE = './resources/glove/glove.840B.300d.txt'
		f = open(EMBEDDING_FILE)
		embeddings_index = dict()
		for line in f:
			#split up line into an indexed array
			values = line.split(' ')
			#first index is word
			word = values[0]
			#store the rest of the values in the array as a new array
			
			coefs = np.asarray(values[1:], dtype='float32')
			embeddings_index[word] = coefs #50 dimensions
		f.close()

		all_embs = np.stack(embeddings_index.values())
		emb_mean,emb_std = all_embs.mean(), all_embs.std()
		embed_size = all_embs.shape[1]

		embedding_matrix_1 = np.random.normal(emb_mean, emb_std, (len(word_index), embed_size))
		for word, i in word_index.items():
			embedding_vector = embeddings_index.get(word)
			if embedding_vector is not None: 
				embedding_matrix_1[i] = embedding_vector

		del embeddings_index; gc.collect()
		embeddings.append({"type": "glove", "embds": embedding_matrix_1, "embd_dim": embed_size}) 

	if "word2vec" in types:
		EMBEDDING_FILE = './resources/GoogleNews-vectors-negative300.bin'
		embeddings_index = dict()
		word2vecDict = word2vec.KeyedVectors.load_word2vec_format("./resources/GoogleNews-vectors-negative300.bin", binary=True)
		for word in word2vecDict.wv.vocab:
			embeddings_index[word] = word2vecDict.word_vec(word)
		embed_size = 300
		
		all_embs = np.stack(embeddings_index.values())
		emb_mean,emb_std = all_embs.mean(), all_embs.std()
		embed_size = all_embs.shape[1]

		embedding_matrix_2 = np.random.normal(emb_mean, emb_std, (len(word_index), embed_size))
		for word, i in word_index.items():
			embedding_vector = embeddings_index.get(word)
			if embedding_vector is not None: 
				embedding_matrix_2[i] = embedding_vector

		del embeddings_index; gc.collect()
		embeddings.append({"type": "word2vec", "embds": embedding_matrix_2, "embd_dim": 300}) 

	if "urban_glove" in types:
		EMBEDDING_FILE = './resources/urban_glove300.txt'
		f = open(EMBEDDING_FILE)
		embeddings_index = dict()
		for line in f:
			#split up line into an indexed array
			values = line.split(' ')
			#first index is word
			word = values[0]
			#store the rest of the values in the array as a new array
			
			coefs = np.asarray(values[1:], dtype='float32')
			embeddings_index[word] = coefs #50 dimensions
		f.close()
		embd_size = 300
		embedding_matrix_3 = np.zeros((len(word_index), embd_size))
		for word, i in word_index.items():
			embedding_vector = embeddings_index.get(word)
			if embedding_vector is not None: embedding_matrix_3[i] = embedding_vector

		del embeddings_index; gc.collect()
		embeddings.append({"type": "urban_glove", "embds": embedding_matrix_3, "embd_dim": embd_size})

	embedding_matrix = np.concatenate([i["embds"] for i in embeddings], axis=1)
	embd_dim = np.sum([i["embd_dim"] for i in embeddings])
	del embeddings
	gc.collect()
	return embedding_matrix, embd_dim

if __name__ == '__main__':
	load_embeddings()
